﻿// DomainServiceProxy.cs
//

using System.Collections.Generic;
using System.Security.Principal;

namespace System.Web.DomainServices
{
    public static class DomainServiceProxy
    {
        private static readonly IDictionary<Type, Type> ProxyTypeMap = new Dictionary<Type, Type>();

        //public static TService Create<TServiceInterface, TService>() where TService : DomainService, TServiceInterface
        //{
        //    throw new NotImplementedException();
        //}

        public static TService Create<TService>() where TService : DomainService
        {
            return Create<TService>(new HttpContextWrapper(HttpContext.Current));
        }

        public static TService Create<TService>(HttpContextBase httpContext) where TService : DomainService
        {
            ServiceProvider serviceProvider = new ServiceProvider(httpContext);
            DomainServiceContext context = new DomainServiceContext(serviceProvider, DomainOperationType.Query);

            return DomainServiceProxy.Create<TService>(context);
        }

        public static TService Create<TService>(DomainServiceContext context) where TService : DomainService
        {
            Type serviceType = typeof(TService);

            if (typeof(IDomainServiceProxy).IsAssignableFrom(serviceType))
            {
                throw new InvalidOperationException("The specified service type already represents a proxy type.");
            }

            if (serviceType.IsSealed || (serviceType.IsPublic == false))
            {
                throw new InvalidOperationException("The specified service type must be public and should not be sealed.");
            }

            // Description-creation performs additional validation
            DomainServiceDescription description = DomainServiceDescription.GetDescription(serviceType);

            Type proxyType = GetProxyType<TService>();
            return (TService)DomainService.Factory.CreateDomainService(proxyType, context);
        }

        public static Type GetProxyType<TService>() where TService : DomainService
        {
            Type serviceType = typeof(TService);
            Type[] serviceTypeArguments = null;

            if (serviceType.IsGenericType)
            {
                serviceTypeArguments = serviceType.GetGenericArguments();
                serviceType = serviceType.GetGenericTypeDefinition();
            }

            Type proxyType = GetProxyType(serviceType);
            if (serviceTypeArguments != null)
            {
                proxyType = proxyType.MakeGenericType(serviceType.GetGenericArguments());
            }

            return proxyType;
        }

        private static Type GetProxyType(Type serviceType)
        {
            Type proxyType;
            if (ProxyTypeMap.TryGetValue(serviceType, out proxyType) == false)
            {
                proxyType = DomainServiceProxyGenerator.Generate(serviceType);
                ProxyTypeMap[serviceType] = proxyType;
            }

            return proxyType;
        }

        private sealed class ServiceProvider : IServiceProvider
        {
            private HttpContextBase _httpContext;

            public ServiceProvider(HttpContextBase httpContext)
            {
                _httpContext = httpContext;
            }

            #region IServiceProvider Members
            object IServiceProvider.GetService(Type serviceType)
            {
                if (serviceType == null)
                {
                    throw new ArgumentNullException("serviceType");
                }

                if (serviceType == typeof(IPrincipal))
                {
                    return _httpContext.User;
                }
                if (serviceType == typeof(HttpContextBase))
                {
                    return _httpContext;
                }

                return null;
            }
            #endregion
        }
    }
}
