﻿// DomainServiceProxyGenerator.cs
//

//#define EMIT_DYNAMIC_ASSEMBLY

using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Linq;
using System.Reflection;
using System.Reflection.Emit;
using ReflectionAssemblyBuilder = System.Reflection.Emit.AssemblyBuilder;
using ReflectionModuleBuilder = System.Reflection.Emit.ModuleBuilder;

namespace System.Web.DomainServices
{
    /// <summary>
    /// Used to generate <see cref="DomainService"/> proxies.
    /// </summary>
    internal static class DomainServiceProxyGenerator
    {
        /// <summary>
        /// Assembly suffix used when defining dynamic assemblies and modules.
        /// </summary>
        private const string ProxyAssemblySuffix = ".{DynamicProxies}";

        /// <summary>
        /// Type suffix used when defining dynamic proxy types.
        /// </summary>
        private const string ProxyTypeSuffix = "{Proxy}";

        /// <summary>
        /// Name of private forward field used to forward method invocations.
        /// </summary>
        private const string ProxyForwardField = "{forwardCount}";

        /// <summary>
        /// Used to construct a dynamic assembly containing dynamic proxy types.
        /// </summary>
        private static ReflectionAssemblyBuilder assemblyBuilder;

        /// <summary>
        /// Used to construct a dynamic module containing dynamic proxy types.
        /// </summary>
        private static ReflectionModuleBuilder moduleBuilder;

        /// <summary>
        /// Static constructor.
        /// </summary>
        static DomainServiceProxyGenerator()
        {
            string assemblyName = typeof(DomainService).Namespace + ProxyAssemblySuffix;

#if EMIT_DYNAMIC_ASSEMBLY
            // Initializes the AssemblyBuilder and ModuleBuilder for debugging purposes. 
            // PDB information will be emitted into the module and the dynamic assembly will be initialized with
            // save-to-disk access permissions.
            assemblyBuilder = AppDomain.CurrentDomain.DefineDynamicAssembly(new AssemblyName(assemblyName), AssemblyBuilderAccess.RunAndSave);
            moduleBuilder = assemblyBuilder.DefineDynamicModule(assemblyName, assemblyName + ".dll", true);
#else
            // Initializes the AssemblyBuilder and ModuleBuilder for release mode.  The assembly builder
            // will only have run access permission and no PDB information will be emitted.
            assemblyBuilder = AppDomain.CurrentDomain.DefineDynamicAssembly(new AssemblyName(assemblyName), AssemblyBuilderAccess.Run);
            moduleBuilder = assemblyBuilder.DefineDynamicModule(assemblyName);
#endif
        }

        /// <summary>
        /// Generates a <see cref="DomainService"/> proxy.
        /// </summary>
        /// <param name="domainServiceType">The <see cref="DomainService"/> type to generate a proxy for.</param>
        /// <returns>A <see cref="DomainService"/> proxy.</returns>
        public static Type Generate(Type domainServiceType)
        {
            Debug.Assert(domainServiceType != null);

            // Define the proxy type
            TypeBuilder typeBuilder = DefineProxyType(domainServiceType);

            // Implement IDomainServiceProxy
            ImplementIDomainServiceProxyInterface(typeBuilder, domainServiceType);

            // Generate proxy type methods (ex: GetCustomers(...))
            DefineOperationOverloads(typeBuilder, domainServiceType);

            //// Generate domain service overrides (Submit, Invoke, Query)
            //DefineBaseOverloads(typeBuilder, domainServiceType);

            // Generate constructors
            DefineConstructors(typeBuilder, domainServiceType);

            // If you build it, They will come.
            Type proxyType = typeBuilder.CreateType();

#if EMIT_DYNAMIC_ASSEMBLY
            // Save the assembly to disk
            SaveAssembly();
#endif

            return proxyType;
        }

        /// <summary>
        /// Defines a proxy type.
        /// </summary>
        private static TypeBuilder DefineProxyType(Type domainServiceType)
        {
            TypeBuilder typeBuilder;

            // Emit:
            //
            //     [ $(Attributes) ]
            //     public class $(DomainServiceTypeName){Proxy} : $(DomainServiceTypeName), IDomainServiceProxy
            //     {
            typeBuilder = moduleBuilder.DefineType(domainServiceType.Name + ProxyTypeSuffix, TypeAttributes.Public);
            typeBuilder.SetParent(domainServiceType);
            typeBuilder.AddInterfaceImplementation(typeof(IDomainServiceProxy));

            // Apply type attributes found in base
            foreach (CustomAttributeData cad in CustomAttributeData.GetCustomAttributes(domainServiceType))
            {
                CustomAttributeBuilder cab = CreateCustomAttributeBuilder(cad);
                typeBuilder.SetCustomAttribute(cab);
            }

            return typeBuilder;
        }

        private static void ImplementIDomainServiceProxyInterface(TypeBuilder typeBuilder, Type domainServiceType)
        {
            // Implement IDomainServiceProxy
            // - Type DomainServiceType { get; }
            // - Dictionary<object, object> CurrentOriginalEntityMap { get; }

            // Emit:
            //
            //      public Type DomainServiceType;  (property definition)
            //
            //      public Type get_DomainServiceType();  (property getter)
            PropertyBuilder typeProperty = typeBuilder.DefineProperty("DomainServiceType", PropertyAttributes.None, typeof(Type), Type.EmptyTypes);
            MethodBuilder typeGetter = typeBuilder.DefineMethod("get_DomainServiceType", MethodAttributes.Public | MethodAttributes.SpecialName | MethodAttributes.HideBySig | MethodAttributes.Virtual, typeof(Type), Type.EmptyTypes);
            ILGenerator typeGetterIL = typeGetter.GetILGenerator();

            // Emit:
            //
            //      public Type get_DomainServiceType()
            //      {
            //          return $(DomainServiceType);
            //      }
            typeGetterIL.Emit(OpCodes.Ldtoken, domainServiceType);
            typeGetterIL.Emit(OpCodes.Call, typeof(Type).GetMethod("GetTypeFromHandle", BindingFlags.Public | BindingFlags.Static));
            typeGetterIL.Emit(OpCodes.Ret);

            // Associate the getter
            typeProperty.SetGetMethod(typeGetter);

            // Emit:
            // 
            //     private Dictionary<object, object> _currentOriginalEntityMap;
            FieldBuilder mapField = typeBuilder.DefineField("_currentOriginalEntityMap", typeof(Dictionary<object, object>), FieldAttributes.Private);

            // Emit:
            //
            //      Dictionary<object, object> CurrentOriginalEntityMap;  (property definition)
            //
            //      Dictionary<object, object> get_CurrentOriginalEntityMap();  (property getter)
            PropertyBuilder mapProperty = typeBuilder.DefineProperty("CurrentOriginalEntityMap", PropertyAttributes.None, typeof(Dictionary<object, object>), Type.EmptyTypes);
            MethodBuilder mapGetter = typeBuilder.DefineMethod("get_CurrentOriginalEntityMap", MethodAttributes.Public | MethodAttributes.SpecialName | MethodAttributes.HideBySig | MethodAttributes.Virtual, typeof(Dictionary<object, object>), Type.EmptyTypes);
            ILGenerator mapGetterIL = mapGetter.GetILGenerator();

            // Emit:
            //
            //      public Dictionary<object, object> get_CurrentOriginalEntityMap()
            //      {
            //          if (this._currentOriginalEntityMap != null)
            //          {
            //              goto returnValue;
            //          }
            //
            //          this._currentOriginalEntityMap = new Dictionary<object, object>();
            //          
            //          returnValue:
            //          return this._currentOriginalEntityMap;
            //      }

            Label returnValue = mapGetterIL.DefineLabel();

            // Emit:
            //
            //      if (this._currentOriginalEntityMap != null)
            //      {
            //          goto returnValue;
            //      }
            mapGetterIL.Emit(OpCodes.Ldarg_0);
            mapGetterIL.Emit(OpCodes.Ldfld, mapField);
            mapGetterIL.Emit(OpCodes.Brtrue, returnValue);

            // Emit:
            //
            //      this._currentOriginalEntityMap = new Dictionary<object, object>();
            mapGetterIL.Emit(OpCodes.Ldarg_0);
            mapGetterIL.Emit(OpCodes.Newobj, typeof(Dictionary<object, object>).GetConstructor(Type.EmptyTypes));
            mapGetterIL.Emit(OpCodes.Stfld, mapField);

            // Emit:
            //
            //      return this._currentOriginalEntityMap;
            mapGetterIL.MarkLabel(returnValue);
            mapGetterIL.Emit(OpCodes.Ldarg_0);
            mapGetterIL.Emit(OpCodes.Ldfld, mapField);
            mapGetterIL.Emit(OpCodes.Ret);

            // Associate the getter
            mapProperty.SetGetMethod(mapGetter);
        }

        /// <summary>
        /// Define proxy constructors.
        /// </summary>
        /// <param name="context">The current generation context.</param>
        private static void DefineConstructors(TypeBuilder typeBuilder, Type domainServiceType)
        {
            ConstructorInfo[] constructors = domainServiceType.GetConstructors();
            foreach (ConstructorInfo constructor in constructors)
            {
                ParameterInfo[] parameters = constructor.GetParameters();
                Type[] parameterTypes = parameters.Select(p => p.IsOut || p.ParameterType.IsByRef ? p.ParameterType.GetElementType() : p.ParameterType).ToArray();

                ConstructorBuilder constructorBuilder =
                    typeBuilder.DefineConstructor(
                        constructor.Attributes,
                        CallingConventions.HasThis,
                        parameterTypes);

                InitializeConstructor(constructorBuilder, constructor);

                ILGenerator methodBody = constructorBuilder.GetILGenerator();

                // Emit:
                //
                //     $(ConstructorAccessModifiers) $(ProxyName) ( [$(ParameterAttributes)] $(Parameters))
                //         : base( $(Parameters) )
                //     {
                //     }
                for (int i = 0; i <= parameters.Length; ++i)
                {
                    methodBody.Emit(OpCodes.Ldarg_S, i);
                }
                methodBody.Emit(OpCodes.Call, constructor);
                methodBody.Emit(OpCodes.Ret);
            }
        }

        /// <summary>
        /// Defines overloads for <see cref="DomainService"/> operations.
        /// </summary>
        private static void DefineOperationOverloads(TypeBuilder typeBuilder, Type domainServiceType)
        {
            DomainServiceDescription domainServiceDescription = DomainServiceDescription.GetDescription(domainServiceType);

            // Iterate through all operations
            foreach (DomainOperationEntry domainOperationEntry in domainServiceDescription.DomainOperationEntries)
            {
                MethodInfo method = domainServiceType.GetMethod(domainOperationEntry.Name, BindingFlags.Public | BindingFlags.Instance);

                // Does the method exist?  We don't support virtualized operations.
                if (method == null)
                {
                    continue;
                }

                // Is the method virtual?
                if (method.IsVirtual == false)
                {
                    throw new InvalidOperationException(
                        string.Format(CultureInfo.CurrentCulture,
                                      "The method named '{0}' on domain service '{1}' is non-virtual. The method must be virtual if it is to be used via a DomainService proxy.",
                                      domainOperationEntry.Name,
                                      domainServiceType.FullName));
                }

                switch (domainOperationEntry.Operation)
                {
                    // Do nothing for Resolve methods, let them pass through.
                    case DomainOperation.Resolve:
                        break;

                    // Generate CUD and Custom methods
                    case DomainOperation.Delete:
                    case DomainOperation.Insert:
                    case DomainOperation.Update:
                    case DomainOperation.Custom:
                        GenerateSubmitMethodOverride(typeBuilder, method, domainOperationEntry.Operation, typeof(DomainServiceProxyHelpers).GetMethod("Submit", BindingFlags.Public | BindingFlags.Static));
                        break;

                    // Generate Query methods
                    case DomainOperation.Query:
                        GenerateQueryOrServiceOperationOverride(typeBuilder, method, typeof(DomainServiceProxyHelpers).GetMethod("Query", BindingFlags.Public | BindingFlags.Static), true);
                        break;

                    // Generate ServiceOperation methods
                    case DomainOperation.ServiceOperation:
                        GenerateQueryOrServiceOperationOverride(typeBuilder, method, typeof(DomainServiceProxyHelpers).GetMethod("Invoke", BindingFlags.Public | BindingFlags.Static), false);
                        break;

                    default:
                        Debug.Fail("Unknown DomainOperation type.");
                        break;
                }
            }
        }

        private static void GenerateQueryOrServiceOperationOverride(TypeBuilder proxyBuilder, MethodInfo method, MethodInfo helperMethod,bool isQueryOperation)
        {
            /* Here's the C# equivalent of what we want to generate:
             * 
             *     // If the helper method is non-void returning service operation or ienumerable returning query
             *     [$(MethodAttributes)]
             *     public override $(ReturnType) $(MethodName)([$(ParameterAttributes)] $(Parameters) )
             *     {
             *         $(ReturnType) result = ($(ReturnType))$(helperMethod)(this, "$(MethodName)", new object[] { $(Parameters) });
             *         return result;
             *     }
             * 
             *     // If the helper method is singleton returning query
             *     [$(MethodAttributes)]
             *     public override $(ReturnType) $(MethodName)([$(ParameterAttributes)] $(Parameters) )
             *     {
             *         $(ReturnType)[] result = ($(ReturnType))$(helperMethod)(this, "$(MethodName)", new object[] { $(Parameters) });
             *         
             *          if (result != null)
             *          {
             *              return result[0];
             *          }
             *          
             *          return null;
             *     }
             * 
             *     // If the helper method is void returning service operation
             *     [$(MethodAttributes)]
             *     public override void $(MethodName)([$(ParameterAttributes)] $(Parameters) )
             *     {
             *         ($(ReturnType))$(helperMethod)(this, "$(MethodName)", new object[] { $(Parameters) });
             *     }
             * 
             */

            ParameterInfo[] parameters = method.GetParameters();
            Type[] parameterTypes = parameters.Select(p => p.IsOut || p.ParameterType.IsByRef ? p.ParameterType.GetElementType() : p.ParameterType).ToArray();
            MethodBuilder queryMethod =
                proxyBuilder.DefineMethod(
                    method.Name,
                    MethodAttributes.Public | MethodAttributes.HideBySig | MethodAttributes.Virtual,
                    method.ReturnType,
                    parameters.Select(pi => pi.ParameterType).ToArray());

            InitializeMethod(queryMethod, method);

            ILGenerator methodBody = queryMethod.GetILGenerator();
            bool voidReturningMethod = method.ReturnType == typeof(void);
            bool isSingletonQuery = isQueryOperation && !typeof(IEnumerable).IsAssignableFrom(method.ReturnType);

            Type resultType;
            if (isSingletonQuery)
            {
                resultType = method.ReturnType.MakeArrayType();
            }
            else if (voidReturningMethod)
            {
                resultType = typeof(object);
            }
            else
            {
                resultType = method.ReturnType;
            }

            // Emit:
            //
            //     $(ReturnType)[] result;
            //     object[] parameterValues;
            LocalBuilder resultLocal = methodBody.DeclareLocal(resultType);
            LocalBuilder paramValuesLocal = methodBody.DeclareLocal(typeof(object[]));

            // Emit:
            //
            //     // For all 'out' parameters
            //     <parameter> = default();
            for (int i = 0; i < parameters.Length; ++i)
            {
                ParameterInfo parameter = parameters[i];

                if (parameter.IsOut && !parameter.IsIn)
                {
                    methodBody.Emit(OpCodes.Ldarg_S, i + 1); // <parameter>
                    methodBody.Emit(OpCodes.Initobj, parameterTypes[i]); // <parameter> = default();
                }
            }

            // Emit:
            //
            //     parameterValues = new object[ $(Parameters).Length ];
            methodBody.Emit(OpCodes.Ldc_I4_S, parameters.Length);
            methodBody.Emit(OpCodes.Newarr, typeof(object));
            methodBody.Emit(OpCodes.Stloc, paramValuesLocal);

            // Emit:
            //
            //     // Repeat for all parameter values...
            //     parameterValues[n] = $(Parameters)[n];
            for (int i = 0; i < parameters.Length; i++)
            {
                ParameterInfo param = parameters[i];

                methodBody.Emit(OpCodes.Ldloc, paramValuesLocal); // parameterValues
                methodBody.Emit(OpCodes.Ldc_I4_S, i);
                methodBody.Emit(OpCodes.Ldarg_S, i + 1); // parameterValues[n]

                // out/ref types?  push on stack
                if (param.IsOut || param.ParameterType.IsByRef)
                {
                    methodBody.Emit(OpCodes.Ldobj, parameterTypes[i]);
                }

                // value or generics?  box it up
                if (parameterTypes[i].IsValueType || parameterTypes[i].IsGenericParameter)
                {
                    methodBody.Emit(OpCodes.Box, parameterTypes[i]);
                }

                methodBody.Emit(OpCodes.Stelem_Ref); // parameterValues[n] = arg[n]
            }

            // Emit:
            // 
            //     result = ($(ReturnType))$(helperMethod)(this, "$(MethodName)", parameterValues);
            methodBody.Emit(OpCodes.Ldarg_0);
            methodBody.Emit(OpCodes.Ldstr, method.Name);
            methodBody.Emit(OpCodes.Ldloc_S, paramValuesLocal);
            methodBody.Emit(OpCodes.Call, helperMethod);

            if (!voidReturningMethod && (method.ReturnType.IsValueType || method.ReturnType.IsGenericParameter))
            {
                methodBody.Emit(OpCodes.Unbox_Any, resultType);
            }
            else
            {
                methodBody.Emit(OpCodes.Castclass, resultType);
            }

            methodBody.Emit(OpCodes.Stloc_S, resultLocal);

            if (!voidReturningMethod)
            {
                if (isSingletonQuery)
                {
                    // Emit:
                    //
                    //     if (result != null)
                    //     {
                    //         return result[0];
                    //     }
                    Label returnArray0 = methodBody.DefineLabel();
                    methodBody.Emit(OpCodes.Ldloc, resultLocal);
                    methodBody.Emit(OpCodes.Brfalse, returnArray0);
                    methodBody.Emit(OpCodes.Ldloc, resultLocal);
                    methodBody.Emit(OpCodes.Ldc_I4_0);
                    methodBody.Emit(OpCodes.Ldelem_Ref);
                    methodBody.Emit(OpCodes.Ret);

                    methodBody.MarkLabel(returnArray0);
                }

                // Emit:
                //
                //     return result;
                methodBody.Emit(OpCodes.Ldloc, resultLocal);
            }

            methodBody.Emit(OpCodes.Ret);
        }

        private static void GenerateSubmitMethodOverride(TypeBuilder proxyBuilder, MethodInfo method, DomainOperation domainOperation, MethodInfo helperMethod)
        {
            /* Here's the C# equivalent of what we want to generate:
            * 
            *     [$(MethodAttributes)]
            *     public override void $(MethodName)($(EntityParameter), (... [$(ParameterAttributes)] $(Parameters)))
            *     {
            *         DomainServiceProxyHelpers.Submit(this, $(EntityParameter), new object[] { $(Parameters) }, $(DomainOperation));
            *     }
            * 
            */

            ParameterInfo[] parameters = method.GetParameters();
            Type[] parameterTypes = parameters.Select(p => p.IsOut || p.ParameterType.IsByRef ? p.ParameterType.GetElementType() : p.ParameterType).ToArray();
            MethodBuilder queryMethod =
                proxyBuilder.DefineMethod(
                    method.Name,
                    MethodAttributes.Public | MethodAttributes.HideBySig | MethodAttributes.Virtual,
                    method.ReturnType,
                    parameters.Select(pi => pi.ParameterType).ToArray());

            InitializeMethod(queryMethod, method);

            ILGenerator methodBody = queryMethod.GetILGenerator();

            // Emit:
            //
            //     object[] parameterValues;
            LocalBuilder paramValuesLocal = methodBody.DeclareLocal(typeof(object[]));

            // Emit:
            //
            //     // For all 'out' parameters
            //     <parameter> = default();
            for (int i = 0; i < parameters.Length; ++i)
            {
                ParameterInfo parameter = parameters[i];

                if (parameter.IsOut && !parameter.IsIn)
                {
                    methodBody.Emit(OpCodes.Ldarg_S, i + 1); // <parameter>
                    methodBody.Emit(OpCodes.Initobj, parameterTypes[i]); // <parameter> = default();
                }
            }

            // Do we need to pass parameters along?  For a submit method, we can 
            // assume the first parameter is the entity so we check to see if
            // additional parameters exist beyond that.
            if ((parameters.Length - 1) < 1)
            {
                // Emit:
                //
                //     parameterValues = null;
                methodBody.Emit(OpCodes.Ldnull);
                methodBody.Emit(OpCodes.Stloc, paramValuesLocal);
            }
            else
            {
                // Emit:
                //
                //     parameterValues = new object[ $(Parameters).Length ];
                methodBody.Emit(OpCodes.Ldc_I4_S, parameters.Length - 1);
                methodBody.Emit(OpCodes.Newarr, typeof(object));
                methodBody.Emit(OpCodes.Stloc, paramValuesLocal);

                // Emit:
                //
                //     // Repeat for all parameter values...
                //     parameterValues[n] = $(Parameters)[n];
                for (int i = 1; i < parameters.Length; i++)
                {
                    ParameterInfo param = parameters[i];

                    methodBody.Emit(OpCodes.Ldloc, paramValuesLocal); // parameterValues
                    methodBody.Emit(OpCodes.Ldc_I4_S, i - 1);
                    methodBody.Emit(OpCodes.Ldarg_S, i + 1); // parameterValues[n]

                    // out/ref types?  push on stack
                    if (param.IsOut || param.ParameterType.IsByRef)
                    {
                        methodBody.Emit(OpCodes.Ldobj, parameterTypes[i]);
                    }

                    // value or generics?  box it up
                    if (parameterTypes[i].IsValueType || parameterTypes[i].IsGenericParameter)
                    {
                        methodBody.Emit(OpCodes.Box, parameterTypes[i]);
                    }

                    methodBody.Emit(OpCodes.Stelem_Ref); // parameterValues[n] = arg[n]
                }
            }

            // Emit:
            // 
            //     $(helperMethod)(this, $(EntityParameter), $(DomainOperation), new object[] { $(Parameters) }, $(DomainOperation));
            methodBody.Emit(OpCodes.Ldarg_0);
            methodBody.Emit(OpCodes.Ldarg_1);
            if (domainOperation == DomainOperation.Custom)
            {
                methodBody.Emit(OpCodes.Ldstr, method.Name);
            }
            else
            {
                methodBody.Emit(OpCodes.Ldnull);
            }
            methodBody.Emit(OpCodes.Ldloc_S, paramValuesLocal);
            methodBody.Emit(OpCodes.Ldc_I4_S, (int)domainOperation);
            methodBody.Emit(OpCodes.Call, helperMethod);
            methodBody.Emit(OpCodes.Ret);
        }

        /// <summary>
        /// Generates overrides for the Invoke, Query and Submit methods.
        /// </summary>
        private static void DefineBaseOverloads(TypeBuilder typeBuilder, Type domainServiceType)
        {
            GenerateBaseProxyMethod(typeBuilder, domainServiceType.GetMethod("Invoke", BindingFlags.Public | BindingFlags.Instance));
            GenerateBaseProxyMethod(typeBuilder, domainServiceType.GetMethod("Query", BindingFlags.Public | BindingFlags.Instance));
            GenerateBaseProxyMethod(typeBuilder, domainServiceType.GetMethod("Submit", BindingFlags.Public | BindingFlags.Instance));
        }

        /// <summary>
        /// Generates an override suitable for use with the Invoke, Query and Submit methods.
        /// </summary>
        private static MethodBuilder GenerateBaseProxyMethod(TypeBuilder proxyBuilder, MethodInfo method)
        {
            /* Here's the C# equivalent of what we want to generate:
             * 
             *     // For a void-returning method:
             *     [$(MethodAttributes)]
             *     public override $(ReturnType) $(MethodName)([$(ParameterAttributes)] $(Params))
             *     {
             *         base.$(MethodName)( $(Params) );
             *     }
             * 
             *     // For a non-void method:
             *     [$(MethodAttributes)]
             *     public override $(ReturnType) $(MethodName)([$(ParameterAttributes)] $(Params))
             *     {
             *         return base.$(MethodName)( $(Params) );
             *     }
             * 
             */

            ParameterInfo[] parameters = method.GetParameters();
            MethodBuilder methodBuilder =
                proxyBuilder.DefineMethod(
                    method.Name,
                    MethodAttributes.Public | MethodAttributes.HideBySig | MethodAttributes.Virtual,
                    method.ReturnType,
                    parameters.Select(pi => pi.ParameterType).ToArray());


            // Initialize our method and parameter attributes
            InitializeMethod(methodBuilder, method);

            // Explicitly add an [IgnoreOperation] attribute to our base overrides
            methodBuilder.SetCustomAttribute(new CustomAttributeBuilder(typeof(IgnoreOperationAttribute).GetConstructor(new Type[] { }), new object[] { }));

            bool voidReturningMethod = method.ReturnType == typeof(void);
            LocalBuilder returnTypeBuilder = null;

            // Start generating the method IL...
            ILGenerator methodBody = methodBuilder.GetILGenerator();

            // Create a local return type if needed
            if (!voidReturningMethod)
            {
                // Emit: 
                // 
                //     $(ReturnType) returnValue;
                returnTypeBuilder = methodBody.DeclareLocal(method.ReturnType);
            }

            // Emit: 
            // 
            //     // If we're returning a value
            //     returnValue = base.$(MethodName)( $(Params) );
            // 
            //     // If the method is returning void
            //     base.$(MethodName)( $(Params) );
            for (int i = 0; i <= parameters.Length; ++i)
            {
                methodBody.Emit(OpCodes.Ldarg_S, i);
            }
            methodBody.Emit(OpCodes.Call, method);
            if (!voidReturningMethod)
            {
                methodBody.Emit(OpCodes.Stloc, returnTypeBuilder);
            }

            // Emit: 
            // 
            //     // If we're returning a value
            //     return returnValue;
            // 
            //     // If the method is returning void
            //     return;
            if (!voidReturningMethod)
            {
                methodBody.Emit(OpCodes.Ldloc, returnTypeBuilder);
            }
            methodBody.Emit(OpCodes.Ret);

            return methodBuilder;
        }

        /// <summary>
        /// Initializers a constructor.
        /// </summary>
        /// <param name="constructorBuilder">The current constructor builder.</param>
        /// <param name="constructorInfo">The <see cref="ConstructorInfo"/> associated with the <paramref name="constructorBuilder"/>.</param>
        private static void InitializeConstructor(ConstructorBuilder constructorBuilder, ConstructorInfo constructorInfo)
        {
            // Apply method attributes
            foreach (CustomAttributeData cad in CustomAttributeData.GetCustomAttributes(constructorInfo))
            {
                CustomAttributeBuilder cab = CreateCustomAttributeBuilder(cad);
                constructorBuilder.SetCustomAttribute(cab);
            }

            // Build parameters
            var parameters = constructorInfo.GetParameters();
            for (int i = 0; i < parameters.Length; ++i)
            {
                ParameterInfo parameterInfo = parameters[i];

                string parameterName = string.IsNullOrEmpty(parameterInfo.Name) ?
                    "param" + i.ToString(CultureInfo.InvariantCulture) :
                    parameterInfo.Name;

                ParameterBuilder paramBuilder =
                    constructorBuilder.DefineParameter(
                        i + 1, // 1-based index!  (0 == return type)
                        parameterInfo.Attributes,
                        parameterName);

                // Apply parameter attributes
                foreach (CustomAttributeData cad in CustomAttributeData.GetCustomAttributes(parameterInfo))
                {
                    CustomAttributeBuilder cab = CreateCustomAttributeBuilder(cad);
                    paramBuilder.SetCustomAttribute(cab);
                }
            }
        }

        /// <summary>
        /// Initializes a method.
        /// </summary>
        /// <param name="methodBuilder">The current constructor builder.</param>
        /// <param name="methodInfo">The <see cref="MethodInfo"/> associated with the <paramref name="methodBuilder"/>.</param>
        private static void InitializeMethod(MethodBuilder methodBuilder, MethodInfo methodInfo)
        {
            // Apply method attributes
            foreach (CustomAttributeData cad in CustomAttributeData.GetCustomAttributes(methodInfo))
            {
                CustomAttributeBuilder cab = CreateCustomAttributeBuilder(cad);
                methodBuilder.SetCustomAttribute(cab);
            }

            // Build parameters
            var parameters = methodInfo.GetParameters();
            for (int i = 0; i < parameters.Length; ++i)
            {
                ParameterInfo parameterInfo = parameters[i];

                string parameterName = string.IsNullOrEmpty(parameterInfo.Name) ?
                    "param" + i.ToString(CultureInfo.InvariantCulture) :
                    parameterInfo.Name;

                ParameterBuilder paramBuilder =
                    methodBuilder.DefineParameter(
                        i + 1, // 1-based index!  (0 == return type)
                        parameterInfo.Attributes,
                        parameterName);

                // Apply parameter attributes
                foreach (CustomAttributeData cad in CustomAttributeData.GetCustomAttributes(parameterInfo))
                {
                    CustomAttributeBuilder cab = CreateCustomAttributeBuilder(cad);
                    paramBuilder.SetCustomAttribute(cab);
                }
            }
        }

        /// <summary>
        /// Creates a <see cref="CustomAttributeBuilder"/> with the provided <see cref="CustomAttributeData"/>.
        /// </summary>
        /// <param name="customAttributeData">The <see cref="CustomAttributeData"/> to build.</param>
        /// <returns>A <see cref="CustomAttributeBuilder"/>.</returns>
        private static CustomAttributeBuilder CreateCustomAttributeBuilder(CustomAttributeData customAttributeData)
        {
            IEnumerable<CustomAttributeNamedArgument> properties = customAttributeData.NamedArguments.Where(na => na.MemberInfo is PropertyInfo);
            IEnumerable<CustomAttributeNamedArgument> fields = customAttributeData.NamedArguments.Where(na => na.MemberInfo is FieldInfo);

            return
                new CustomAttributeBuilder(
                    customAttributeData.Constructor, // ctor ref
                    customAttributeData.ConstructorArguments.Select(ca => ca.Value).ToArray(), // ctor values
                    properties.Select(p => p.MemberInfo).Cast<PropertyInfo>().ToArray(), // property refs
                    properties.Select(p => p.TypedValue.Value).ToArray(), // property values
                    fields.Select(f => f.MemberInfo).Cast<FieldInfo>().ToArray(), // field refs
                    fields.Select(f => f.TypedValue.Value).ToArray()); // field values
        }

        /// <summary>
        /// Helper method gets a <see cref="MethodInfo"/> for generic Func or Action delegate.
        /// </summary>
        /// <param name="genericType">The generic delegate type.</param>
        /// <param name="genericTypes">The generic delegate's type arguments.</param>
        /// <returns>A <see cref="MethodInfo"/> that can be used to invoke delegates of type <paramref name="genericType"/>.</returns>
        private static MethodInfo GetDelegateMethodInfo(Type genericType, Type[] genericTypes)
        {
            // Create an array of parameter types for use with Func<> delegates.
            // We're shaving 1 off the array size as the last type argument for a Func<>
            // is always the return type.
            Type[] parameterTypes = new Type[genericTypes.Length - 1];

            // Check for any sort of Func<>.  Assumes that this method is not publicly exposed
            // and that we'll either be passed an Action<?> or Func<?>.
            // Here, we do a simple string check (rather than iterate through all the possible variations)
            // as the input type is constrained.
            if (genericType.Name.StartsWith("Func", StringComparison.Ordinal))
            {
                for (int i = 0; i < parameterTypes.Length; i++)
                {
                    parameterTypes[i] = genericTypes[i];
                }
            }
            else
            {
                // If this is an Action<?>, we'll use the types as-is.
                parameterTypes = genericTypes;
            }

            return genericType.GetMethod("Invoke", BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic, null, parameterTypes, null);
        }

#if EMIT_DYNAMIC_ASSEMBLY
        /// <summary>
        /// This method can be used to save the dynamic assembly to disk. Note that the assembly will be saved 
        /// into the current executing directory using a name similar to "System.Web.DomainServices.{DynamicProxies}.dll".
        /// </summary>
        /// <remarks>
        /// Once the dynamic assembly is written to disk, additional modifications (e.g., creating a new proxy type) will
        /// be prohibited.
        /// </remarks>
        private static void SaveAssembly()
        {
            assemblyBuilder.Save(typeof(DomainService).Namespace + ProxyAssemblySuffix + ".dll");
        }
#endif
    }
}
