﻿// DomainServiceProxyHelpers.cs
//

using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Web.Ria.Data;

namespace System.Web.DomainServices
{
    public static class DomainServiceProxyHelpers
    {
        /// <summary>
        /// Delegate used by proxy types to perform <see cref="DomainService"/> query operations.
        /// </summary>
        /// <param name="domainService">The <see cref="DomainService"/> that is performing a query operation.</param>
        /// <param name="name">The name of the query method to invoke.</param>
        /// <param name="parameters">The query operation parameters.</param>
        /// <returns>An <see cref="IEnumerable"/> return from the query.</returns>
        public static IEnumerable Query(DomainService domainService, string name, object[] parameters)
        {
            if (domainService == null)
            {
                throw new ArgumentNullException("domainService");
            }

            if (string.IsNullOrEmpty(name))
            {
                throw new ArgumentNullException("name");
            }

            IDomainServiceProxy domainServiceProxy = domainService as IDomainServiceProxy;

            if (domainServiceProxy == null)
            {
                throw new ArgumentException("The DomainService proxy must implement IDomainServiceProxy.", "domainService");
            }

            domainService = GetProxyBase(domainServiceProxy);

            int totalCount;
            DomainOperationEntry queryOperation = domainService.Description.GetQueryMethod(name);
            QueryDescription queryDescription = new QueryDescription(queryOperation, new ReadOnlyCollection<object>(parameters));

            return domainService.Query(queryDescription, out totalCount);
        }

        /// <summary>
        /// Delegate used by proxy types to perform <see cref="DomainService"/> submit operations.
        /// </summary>
        /// <param name="domainService">The <see cref="DomainService"/> that is performing a submit operation.</param>
        /// <param name="entity">The entity affected by </param>
        /// <param name="parameters">The submit parameters. This can be used in Custom operations.</param>
        /// <param name="domainOperation">The type of submit operation.</param>
        public static void Submit(DomainService domainService, object entity, string operationName, object[] parameters, DomainOperation domainOperation)
        {
            if (domainService == null)
            {
                throw new ArgumentNullException("domainService");
            }

            if (entity == null)
            {
                throw new ArgumentNullException("entity");
            }

            IDomainServiceProxy domainServiceProxy = domainService as IDomainServiceProxy;

            if (domainServiceProxy == null)
            {
                throw new ArgumentException("The DomainService proxy must implement IDomainServiceProxy.", "domainService");
            }

            domainService = GetProxyBase(domainServiceProxy);

            object originalEntity;
            if (domainServiceProxy.CurrentOriginalEntityMap.TryGetValue(entity, out originalEntity))
            {
                domainServiceProxy.CurrentOriginalEntityMap.Remove(entity);
            }

            EntityOperation entityOperation =
                new EntityOperation()
                {
                    Entity = entity,
                    OriginalEntity = originalEntity,
                    OperationName = operationName,
                    Operation = domainOperation,
                    OperationData = parameters
                };

            ChangeSet changeSet = new ChangeSet(new[] { entityOperation });

            domainService.Submit(changeSet);

            if (entityOperation.HasError)
            {
                // REVIEW: Are we concerned about stack traces revealing the true source of this exception?
                throw new EntityOperationException("An error occurred during processing of the submitted ChangeSet. Inspect the EntityOperation property for more information.", entityOperation);
            }
        }

        /// <summary>
        /// Delegate used by proxy types to perform <see cref="DomainService"/> invoke operations.
        /// </summary>
        /// <param name="domainService">The <see cref="DomainService"/> that is performing a invoke operation.</param>
        /// <param name="name">The name of the invoke method to invoke.</param>
        /// <param name="parameters">The invoke operation parameters.</param>
        /// <returns>An object returned by the invoke operation.</returns>
        public static object Invoke(DomainService domainService, string name, object[] parameters)
        {
            if (domainService == null)
            {
                throw new ArgumentNullException("domainService");
            }

            if (string.IsNullOrEmpty(name))
            {
                throw new ArgumentNullException("name");
            }

            IDomainServiceProxy domainServiceProxy = domainService as IDomainServiceProxy;

            if (domainServiceProxy == null)
            {
                throw new ArgumentException("The DomainService proxy must implement IDomainServiceProxy.", "domainService");
            }

            domainService = GetProxyBase(domainServiceProxy);

            List<OperationError> validationErrors = new List<OperationError>();
            DomainOperationEntry method = domainService.Description.GetServiceOperation(name);

            object result = domainService.Invoke(method, parameters, validationErrors);

            if (validationErrors.Count > 0)
            {
                // REVIEW: Are we concerned about stack traces revealing the true source of this exception?
                // throw new EntityOperationException(
                //     Resource.DomainServiceProxy_ServiceOperationError,
                //     new EntityOperation {
                //         ValidationErrors = validationErrors
                //     });
                throw new Exception();
            }

            return result;
        }

        private static DomainService GetProxyBase(IDomainServiceProxy domainService)
        {
            IDomainServiceProxy domainServiceProxy = domainService as IDomainServiceProxy;

            if (domainServiceProxy == null)
            {
                throw new ArgumentException("The DomainService proxy must implement IDomainServiceProxy.", "domainService");
            }

            Type domainServiceType = domainServiceProxy.DomainServiceType;
            return DomainService.Factory.CreateDomainService(domainServiceType, ((DomainService)domainServiceProxy).ServiceContext);
        }
    }
}
